import pygobbler as pyg
import tempfile
import os
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test-upload", staging=staging, url=url)
    pyg.remove_project("test-more-upload", staging=staging, url=url)
    pyg.remove_project("test-upload-perms", staging=staging, url=url)
    pyg.create_project("test-upload", staging=staging, url=url)
    pyg.create_project("test-more-upload", staging=staging, url=url)
    pyg.create_project("test-upload-perms", staging=staging, url=url)

    tmp = tempfile.mkdtemp()
    with open(os.path.join(tmp, "blah.txt"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(tmp, "foo"))
    with open(os.path.join(tmp, "foo", "bar.txt"), "w") as f:
        f.write("1 2 3 4 5 6 7 8 9 10")

    pyg.upload_directory(
        project="test-upload", 
        asset="jennifer", 
        version="1", 
        directory=tmp,
        staging=staging, 
        url=url
    )

    return tmp


def test_upload_directory_simple(setup):
    _, staging, registry, url = pyg.start_gobbler()

    # Checking that the files were, in fact, correctly uploaded.
    man = pyg.fetch_manifest("test-upload", "jennifer", "1", registry=registry, url=url)
    assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt"]
    for k, v in man.items():
        assert "link" not in v

    # Deduplication happens naturally.
    pyg.upload_directory(
        project="test-upload", 
        asset="jennifer", 
        version="2", 
        directory=setup, # i.e., the 'tmp' returned by the setup.
        staging=staging,
        url=url
    )

    man = pyg.fetch_manifest("test-upload", "jennifer", "2", registry=registry, url=url)
    assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt"]
    for k, v in man.items():
        assert "link" in v


def test_upload_directory_links(setup):
    _, staging, registry, url = pyg.start_gobbler()

    dest = tempfile.mkdtemp()
    pyg.clone_version("test-upload", "jennifer", "2", dest, registry=registry)
    with open(os.path.join(dest, "whee"), "w") as f:
        f.write("BLAH")

    pyg.upload_directory(
        project="test-more-upload", 
        asset="natalie", 
        version="1", 
        directory=dest,
        staging=staging,
        url=url
    )

    man = pyg.fetch_manifest("test-more-upload", "natalie", "1", registry=registry, url=url)
    assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt", "whee"]
    assert "link" in man["blah.txt"]
    assert "link" in man["foo/bar.txt"]
    assert "link" not in man["whee"]


def test_upload_directory_staging(setup):
    _, staging, registry, url = pyg.start_gobbler()

    dir = pyg.allocate_upload_directory(staging)
    with open(os.path.join(dir, "blah.txt"), "w") as f:
        f.write("A B C D E")
    os.mkdir(os.path.join(dir, "foo"))
    with open(os.path.join(dir, "foo", "bar.txt"), "w") as f:
        f.write("1 2 3 4 5 6 7 8 9 10")

    pyg.upload_directory(
        project="test-upload", 
        asset="jennifer", 
        version="3", 
        directory=dir,
        staging=staging,
        url=url
    )

    man  = pyg.fetch_manifest("test-upload", "jennifer", "3", registry=registry, url=url)
    assert sorted(man.keys()) == ["blah.txt", "foo/bar.txt"]
    assert "link" not in man["blah.txt"]
    assert "link" in man["foo/bar.txt"]



import pygobbler as pyg


def test_service_info():
    _, staging, registry, url = pyg.start_gobbler()
    payload = pyg.service_info(url)
    assert payload["staging"] == staging
    assert payload["registry"] == registry



import os
from .fetch_manifest import fetch_manifest


def clone_version(project: str, asset: str, version: str, destination: str, registry: str):
    """
    Clone the directory structure for a versioned asset into a separate location.
    This is typically used to prepare a new version for a lightweight upload.

    More specifically, cloning involves creating a directory at ``destination``
    that has the same structure as that of the specified project-asset-version.
    All files in the version are represented as symlinks from ``destination``
    to the corresponding file in the ``registry``.The idea is that, when
    ``destination`` is used in :py:func:`~.upload_directory`, the symlinks are
    converted into upload links. This allows users to create new versions very
    cheaply as duplicate files are not stored in the backend.

    Users can more-or-less do whatever they want inside the cloned ``destination``,
    but the symlink targets should be read-only as they refer to immutable files in
    the ``registry``.  If a file in ``destination`` needs to be modified, the
    symlink should be deleted and replaced with a new file.

    Args:
        project:
            The name of the project.

        asset:
            The name of the asset of the ``project``.

        version:
            The name of the version of the ``asset`` to clone.

        destination:
            Path to a destination directory at which to create the clone.

        registry: 
            Path to the registry.
    """
    target = os.path.join(registry, project, asset, version)
    listing = fetch_manifest(project, asset, version, registry=registry, url=None)
    os.makedirs(destination, exist_ok=True)
    for x in listing.keys():
        dpath = os.path.join(destination, x)
        os.makedirs(os.path.dirname(dpath), exist_ok=True)
        os.symlink(os.path.join(target, x), dpath)
    return



import os
from . import allocate_upload_directory
from . import _utils as ut


def upload_directory(project: str, asset: str, version: str, directory: str, staging: str, url: str, probation: bool = False):
    """
    Upload a directory as a new versioned asset of a project in the registry.

    Args:
        project:
            The name of an existing project.

        asset:
            The name of a new or existing asset in ``project``.

        version:
            The name of a new version of ``asset``.

        directory:
            Path to a directory to be uploaded. For best performace, this
            should be a subdirectory of ``staging``, e.g., as created by
            :py:func:`~.allocate_upload_directory`.

        staging:
            Path to the staging directory.

        url:
            URL for the Gobbler REST API.

        probation:
            Whether to upload a probational version.
    """
    directory = os.path.normpath(directory)
    staging = os.path.normpath(staging)

    if os.path.dirname(directory) != staging:
        newdir = allocate_upload_directory(staging) 

        for root, dirs, files in os.walk(directory):
            for f in files:
                src = os.path.join(root, f)
                rel = os.path.relpath(src, directory)
                dest = os.path.join(newdir, rel)
                os.makedirs(os.path.dirname(dest), exist_ok=True)

                slink = ""
                if os.path.islink(src):
                    slink = os.readlink(src)

                if slink == "" or not os.path.isabs(slink):
                    try:
                        os.link(src, dest)
                    except:
                        import shutil
                        shutil.copy(src, dest)
                else:
                    os.symlink(slink, dest)

        directory = newdir

    req = {
        "source": os.path.basename(directory),
        "project": project,
        "asset": asset,
        "version": version,
        "on_probation": probation
    }
    ut.dump_request(staging, url, "upload", req)
    return



def allocate_upload_directory(staging: str) -> str:
    """
    Allocate a subdirectory in the staging directory to prepare files for
    upload via :py:func:`~.upload_directory`.

    Args:
        staging:
            Path to the staging directory.

    Returns:
        Path to a new subdirectory for staging uploads.
    """
    import tempfile
    return tempfile.mkdtemp(dir=staging)



from typing import Optional, List
import os
import requests
from . import _utils as ut


def list_files(project: str, asset: str, version: str, registry: str, url: str, prefix: Optional[str] = None , include_dotdot: bool = True, force_remote: bool = False) -> List[str]:
    """
    List the contents of a version of a project asset.

    Args:
        project:
            The name of the project.

        asset:
            The name of the asset in ``project``.

        version:
            The name of the version of the ``asset``.

        registry:
            Path to the registry.

        url:
            URL to the Gobbler REST API. Only used for remote access.

        prefix:
            Prefix for the path within this version's subdirectory. If
            provided, files are only listed if they have a relative path (i.e.,
            inside the version subdirectory) that starts with this prefix. If
            None, all files associated with this version are listed.

        include_dotdot:
            Whether to list files with path components that start with ``..``.

        force_remote:
            Whether to force remote access via the API, even if ``registry`` is
            on the same filesystem as the caller. 

    Returns:
        List of strings containing the relative paths of files associated with
        the versioned asset. All paths will start with ``prefix`` if provided.
    """
    suffix_filter = None
    if prefix is not None:
        if prefix.endswith("/"):
            prefix = prefix[:-1]
        else:
            suffix_filter = os.path.basename(prefix)
            prefix = os.path.dirname(prefix)
            if prefix == "":
                prefix = None

    listing = []

    if not force_remote and os.path.exists(registry):
        target = os.path.join(registry, project, asset, version)
        if prefix is not None:
            target = os.path.join(target, prefix)

        for root, dirs, files in os.walk(target):
            rel = os.path.relpath(root, target)
            for f in files:
                if include_dotdot or not f.startswith(".."):
                    if rel != ".":
                        listing.append(os.path.join(rel, f))
                    else:
                        listing.append(f)

    else:
        target = project + "/" + asset + "/" + version
        if prefix is not None:
            target += "/" + prefix

        import urllib
        res = requests.get(url + "/list?path=" + urllib.parse.quote_plus(target) + "&recursive=true")
        if res.status_code >= 300:
            raise ut.format_error(res)

        body = res.json()
        if include_dotdot:
            listing = body
        else:
            listing = filter(lambda x : not x.startswith("..") and x.find("/..") == -1, body)

    if suffix_filter is not None:
        listing = filter(lambda x : x.startswith(suffix_filter), listing)

    if prefix is not None:
        for i, x in enumerate(listing):
            listing[i] = prefix + "/" + x

    return listing



from . import _utils as ut


def remove_project(project: str, staging: str, url: str):
    """
    Remove a project from the registry.

    Args:
        project:
            String containing the project to remove.

        staging:
            Path to the staging directory.

        url:
            URL for the Gobbler REST API.
    """
    ut.dump_request(staging, url, "delete_project", { "project": project })



from typing import Optional
import os
import tempfile
import urllib
import requests
import shutil
from . import _utils as ut


def _local_registry(cache: Optional[str], url: str) -> str:
    if cache is None:
        import appdirs
        cache = appdirs.user_data_dir("gobbler", "aaron")
    return os.path.join(cache, urllib.parse.quote_plus(url))


def _acquire_file(cache: str, path: str, name: str, url: str, overwrite: bool) -> str:
    target = os.path.join(cache, "REGISTRY", path, name)

    if overwrite or not os.path.exists(target):
        tempdir = os.path.join(cache, "TEMP")
        os.makedirs(tempdir, exist_ok=True)
        os.makedirs(os.path.dirname(target), exist_ok=True)

        tempfid, temppath = tempfile.mkstemp(dir=tempdir)
        try:
            with requests.get(url + "/fetch/" + path + "/" + name, stream=True) as r:
                if r.status_code >= 300:
                    raise ut.format_error(r)
                with os.fdopen(tempfid, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)
            os.rename(temppath, target) # this should be more or less atomic, so no need for locks.
        finally:
            try:
                os.remove(temppath)
            except:
                pass

    return target


def fetch_directory(path: str, registry: str, url: str, cache: Optional[str] = None, force_remote: bool = False, overwrite: bool = False, concurrent: int = 1) -> str:
    """
    Obtain the path to a directory from the registry. This may create a local
    copy of the subdirectory's contents if the caller is not on the same
    filesystem as the registry.

    Args:
        path: 
            Relative path to a subdirectory within the registry. This usually
            takes the form of ``PROJECT/ASSET/VERSION/*``.

        registry:
            Path to the registry.

        url:
            URL to the Gobbler REST API. Only used for remote queries.

        cache:
            Path to a cache directory. If None, an appropriate location is
            automatically chosen. Only used for remote access.

        force_remote:
            Whether to force remote access. This will download all files in the
            ``path`` via the REST API and cache them locally, even if
            ``registry`` is present on the same filesystem.

        overwrite:
            Whether to overwrite existing files in the cache.

        concurrent:
            Number of concurrent downloads.

    Returns:
        Path to the subdirectory on the caller's filesystem.  This is either a
        path to the registry if it is accessible, or a path to a local cache of
        the registry's contents otherwise.
    """
    if not force_remote and os.path.exists(registry):
        return os.path.join(registry, path)

    cache = _local_registry(cache, url)
    final = os.path.join(cache, "REGISTRY", path)
    ok = os.path.join(cache, "SUCCESS", path, "....OK")
    if not overwrite and os.path.exists(ok) and os.path.exists(final):
        return final

    res = requests.get(url + "/list?path=" + urllib.parse.quote_plus(path) + "&recursive=true")
    if res.status_code >= 300:
        raise ut.format_error(res)
    listing = res.json()

    if concurrent == 1:
        for y in listing:
            _acquire_file(cache, name=y, path=path, url=url, overwrite=overwrite)
    else:
        import multiprocessing
        import functools
        with multiprocessing.Pool(concurrent) as p:
            p.map(functools.partial(_acquire_file, cache, path, url=url, overwrite=overwrite), listing)

    # We use a directory-level OK file to avoid having to scan through all 
    # the directory contents to indicate that it's complete.
    os.makedirs(os.path.dirname(ok), exist_ok=True)
    with open(ok, "w") as handle:
        handle.write("")
    return final



from typing import Dict, Any
import requests
from . import _utils as ut


def service_info(url: str) -> Dict[str, Any]:
    """
    Get information about the Gobbler service, namely the locations of the
    staging directory and registry.

    Args:
        url:
            URL of the gobbler REST API.

    Returns:
        Dictionary containing the location of the staging and registry
        directories.
    """
    res = requests.get(url + "/info")
    if res.status_code >= 300:
        raise ut.format_error(res)
    return res.json()



import pygobbler as pyg
import tempfile
import os
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()
    pyg.remove_project("test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)

    tmp = tempfile.mkdtemp()
    with open(os.path.join(tmp, "blah.txt"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(tmp, "foo"))
    with open(os.path.join(tmp, "foo", "bar.txt"), "w") as f:
        f.write("1 2 3 4 5 6 7 8 9 10")

    pyg.upload_directory(
        project="test", 
        asset="serenity", 
        version="firefly", 
        directory=tmp,
        staging=staging, 
        url=url
    )


def test_version_path(setup):
    _, staging, registry, url = pyg.start_gobbler()
    out = pyg.version_path("test", "serenity", "firefly", registry=registry, url=url)
    assert out == os.path.join(registry, "test", "serenity", "firefly")
    with open(os.path.join(out, "blah.txt"), "r") as handle:
        assert handle.read() == "BAR"

    cache = tempfile.mkdtemp()
    out = pyg.version_path("test", "serenity", "firefly", registry=registry, url=url, cache=cache, force_remote=True)
    assert out.startswith(cache)
    assert out.endswith("test/serenity/firefly")
    with open(os.path.join(out, "foo", "bar.txt"), "r") as handle:
        assert handle.read() == "1 2 3 4 5 6 7 8 9 10"



from . import fetch_directory


def version_path(project, asset, version, **kwargs) -> str:
    """
    Obtain the path to a versioned asset in the registry.

    Args:
        project:
            The name of the project.

        asset:
            The name of the asset of the ``project``.

        version:
            The name of the version of the ``asset`` to obtain.

        kwargs:
            Further arguments to :py:func:`~.fetch_directory`.

    Returns:
        The path to the directory containing the desired version.
    """
    return fetch_directory(project + "/" + asset + "/" + version, **kwargs)



from typing import Optional
from . import _utils as ut


def refresh_latest(project: str, asset: str, staging: str, url:str) -> Optional[str]:
    """
    Recompute the latest version of a project's asset. This is useful on rare
    occasions where multiple simultaneous uploads cause the latest version to
    be slightly out of sync.

    Args:
        project:
            Name of the project.

        asset:
            Name of the asset.

        staging:
            Path to the staging directory.

        url:
            URL of the gobbler REST API.

    Returns:
        Latest version of the project, or None if there is no non-probational version.
    """
    resp = ut.dump_request(staging, url, "refresh_latest", { "project": project, "asset": asset })
    if "version" in resp:
        return resp["version"]
    else:
        return None



import pygobbler as pyg
import pytest
import os


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()
    pyg.remove_project("test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)

    src = pyg.allocate_upload_directory(staging)
    with open(os.path.join(src, "foo"), "w") as handle:
        handle.write("BAR")
    pyg.upload_directory("test", "probation", "good", src, staging=staging, url=url, probation=True)
    pyg.upload_directory("test", "probation", "bad", src, staging=staging, url=url, probation=True)


def test_approve_probation(setup):
    _, staging, registry, url = pyg.start_gobbler()
    assert pyg.fetch_summary("test", "probation", "good", registry=registry, url=url)["on_probation"]
    assert pyg.fetch_latest("test", "probation", registry=registry, url=url) is None

    pyg.approve_probation("test", "probation", "good", staging=staging, url=url)
    assert "on_probation" not in pyg.fetch_summary("test", "probation", "good", registry=registry, url=url)
    assert pyg.fetch_latest("test", "probation", registry=registry, url=url) == "good"


def test_reject_probation(setup):
    _, staging, registry, url = pyg.start_gobbler()
    assert pyg.fetch_summary("test", "probation", "bad", registry=registry, url=url)["on_probation"]

    pyg.reject_probation("test", "probation", "bad", staging=staging, url=url)
    assert "bad" not in pyg.list_versions("test", "probation", registry=registry, url=url)



from typing import Optional, Dict, Any
import os
import json
from . import fetch_directory as fd


def fetch_summary(project: str, asset: str, version: str, registry: str, url: str, cache: Optional[str] = None, force_remote: bool = False, overwrite: bool = False) -> Dict[str, Any]:
    """
    Fetch the summary for a particular version of a project asset.

    Args:
        project:
            Name of a project.

        asset:
            Name of an asset in the ``project``.

        version:
            Name of a version of the ``asset``.

        registry:
            Path to the Gobbler registry.

        url:
            URL of the REST API. Only used for remote queries.

        cache:
            Path to a cache directory. If None, a default cache location is
            selected. Only used for remote queries.

        force_remote:
            Whether to force a remote query via ``url``, even if the
            ``registry`` is present on the current filesystem.

        overwrite:
            Whether to overwrite existing entries in the cache. Only used for
            remote queries.

    Returns:
        Dictionary containing summary information including:

        - ``uploader_user_id``, string containing the UID of the uploading user.
        - ``upload_start_time``, string containing the upload start time in
          Internet Date/Time format.
        - ``upload_finish_time``, string containing the upload finish time in
          Internet Date/Time format.
        - ``on_probation`` (optional), boolean indicating whether this version of
          the asset is a probational upload.
    """

    if not force_remote and os.path.exists(registry):
        path = os.path.join(registry, project, asset, version, "..summary")
    else:
        cache = fd._local_registry(cache, url)
        path = fd._acquire_file(cache, project + "/" + asset + "/" +  version, "..summary", url=url, overwrite=overwrite)

    with open(path, "r") as handle:
        out = json.load(handle)
    return out



import pygobbler as pyg
import tempfile
import os
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)

    src = pyg.allocate_upload_directory(staging)
    with open(os.path.join(src, "foo"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(src, "whee"))
    with open(os.path.join(src, "whee", "blah"), "w") as f:
        f.write("stuff")

    pyg.upload_directory("test", "fetch", "v1", src, staging=staging, url=url)
    pyg.upload_directory("test", "fetch", "v2", src, staging=staging, url=url)


def test_fetch_manifest(setup):
    _, staging, registry, url = pyg.start_gobbler()

    man = pyg.fetch_manifest("test", "fetch", "v1", registry=registry, url=url)
    assert man["foo"]["size"] == 3
    assert man["whee/blah"]["size"] == 5

    cache = tempfile.mkdtemp()
    rman = pyg.fetch_manifest("test", "fetch", "v1", registry=registry, url=url, cache=cache, force_remote=True)
    assert man == rman


def test_fetch_summary(setup):
    _, staging, registry, url = pyg.start_gobbler()

    summ = pyg.fetch_summary("test", "fetch", "v1", registry=registry, url=url)
    assert isinstance(summ["upload_user_id"], str)
    assert isinstance(summ["upload_start"], str)
    assert isinstance(summ["upload_finish"], str)

    cache = tempfile.mkdtemp()
    rsumm = pyg.fetch_summary("test", "fetch", "v1", registry=registry, url=url, cache=cache, force_remote=True)
    assert summ == rsumm


def test_fetch_latest(setup):
    _, staging, registry, url = pyg.start_gobbler()

    assert pyg.fetch_latest("test", "fetch", registry=registry, url=url) == "v2"
    assert pyg.fetch_latest("test", "missing", registry=registry, url=url) is None
    assert pyg.fetch_latest("test", "fetch", registry=registry, url=url, force_remote=True) == "v2"
    assert pyg.fetch_latest("test", "missing", registry=registry, url=url, force_remote=True) is None


def test_fetch_usage(setup):
    _, staging, registry, url = pyg.start_gobbler()

    assert pyg.fetch_usage("test", registry=registry, url=url) > 0
    assert pyg.fetch_usage("test", registry=registry, url=url, force_remote=True) > 0


def test_fetch_permissions(setup):
    _, staging, registry, url = pyg.start_gobbler()

    perms = pyg.fetch_permissions("test", registry=registry, url=url)
    assert isinstance(perms["owners"], list)
    assert isinstance(perms["uploaders"], list)

    rperms = pyg.fetch_permissions("test", registry=registry, url=url, force_remote=True)
    assert perms == rperms


def test_fetch_directory(setup):
    _, staging, registry, url = pyg.start_gobbler()

    dir = pyg.fetch_directory("test/fetch/v2", registry=registry, url=url)
    assert dir.startswith(registry)
    with open(os.path.join(dir, "foo"), "r") as handle:
        assert handle.read() == "BAR"

    cache = tempfile.mkdtemp()
    rdir = pyg.fetch_directory("test/fetch/v2", registry=registry, url=url, cache=cache, force_remote=True)
    assert rdir.startswith(cache)
    with open(os.path.join(dir, "foo"), "r") as handle:
        assert handle.read() == "BAR"
    with open(os.path.join(dir, "whee", "blah"), "r") as handle:
        assert handle.read() == "stuff"

    # Subsequent requests are no-ops.
    with open(os.path.join(rdir, "foo"), "w") as handle:
        handle.write("more-bar")
    rdir2 = pyg.fetch_directory("test/fetch/v2", registry=registry, url=url, cache=cache, force_remote=True)
    assert rdir == rdir2
    with open(os.path.join(rdir2, "foo"), "r") as handle:
        assert handle.read() == "more-bar"

    # Unless we force an overwrite.
    rdir2 = pyg.fetch_directory("test/fetch/v2", registry=registry, url=url, cache=cache, force_remote=True, overwrite=True)
    with open(os.path.join(rdir2, "foo"), "r") as handle:
        assert handle.read() == "BAR"

    # Trying with multiple cores.
    cache = tempfile.mkdtemp()
    rdir2 = pyg.fetch_directory("test/fetch/v2", registry=registry, url=url, cache=cache, force_remote=True, concurrent=2)
    with open(os.path.join(rdir2, "foo"), "r") as handle:
        assert handle.read() == "BAR"



from typing import Optional, Dict, Any
import os
import json
from . import fetch_directory as fd


def fetch_manifest(project: str, asset: str, version: str, registry: str, url: str, cache: Optional[str] = None, force_remote: bool = False, overwrite: bool = False) -> Dict[str, Any]:
    """
    Fetch the manifest for a version of a project asset.

    Args:
        project:
            Name of a project.

        asset:
            Name of an asset in the ``project``.

        version:
            Name of a version of the ``asset``.

        registry:
            Path to the Gobbler registry.

        url:
            URL of the REST API. Only used for remote queries.

        cache:
            Path to a cache directory. If None, a default cache location is
            selected. Only used for remote queries.

        force_remote:
            Whether to force a remote query via ``url``, even if the
            ``registry`` is present on the current filesystem.

        overwrite:
            Whether to overwrite existing entries in the cache. Only used for
            remote queries.

    Returns:
        Dictionary containing the manifest. Each key is a relative path to a
        file in this version of the project asset, and each value is a
        dictionary with the following fields:

        - ``size``, integer specifying the size of the file in bytes.
        - ``md5sum``, string containing the file's hex-encoded MD5 checksum.
        - ``link`` (optional): a list specifying the link destination for a
          file. This contains the strings ``project``, ``asset``,
          ``version`` and ``path``; if the link destination is also a link,
          an ``ancestor`` dictionary will be present containing the final
          location of the file after resolving all intermediate links. 
    """
    if not force_remote and os.path.exists(registry):
        path = os.path.join(registry, project, asset, version, "..manifest")
    else:
        cache = fd._local_registry(cache, url)
        path = fd._acquire_file(cache, project + "/" + asset + "/" + version, "..manifest", url=url, overwrite=overwrite)

    with open(path, "r") as handle:
        out = json.load(handle)
    return out



import pygobbler as pyg
import os
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test", staging=staging, url=url)
    pyg.remove_project("more-list-test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)
    pyg.create_project("more-list-test", staging=staging, url=url)

    src = pyg.allocate_upload_directory(staging)
    with open(os.path.join(src, "foo"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(src, "whee"))
    with open(os.path.join(src, "whee", "blah"), "w") as f:
        f.write("stuff")
    with open(os.path.join(src, "whee2"), "w") as f:
        f.write("ABCDEFGHIJK")

    pyg.upload_directory("test", "list", "v1", src, staging=staging, url=url)
    pyg.upload_directory("test", "list", "v2", src, staging=staging, url=url)
    pyg.upload_directory("test", "more-list", "foo", src, staging=staging, url=url)
    pyg.upload_directory("more-list-test", "list", "bar", src, staging=staging, url=url)


def test_list_projects(setup):
    _, staging, registry, url = pyg.start_gobbler()

    projects = pyg.list_projects(registry=registry, url=url)
    assert "test" in projects
    assert "more-list-test" in projects

    rprojects = pyg.list_projects(registry=registry, url=url, force_remote=True)
    assert sorted(projects) == sorted(rprojects)


def test_list_assets(setup):
    _, staging, registry, url = pyg.start_gobbler()

    assets = pyg.list_assets("test", registry=registry, url=url)
    assert "list" in assets
    assert "more-list" in assets

    rassets = pyg.list_assets("test", registry=registry, url=url, force_remote=True)
    assert sorted(assets) == sorted(rassets)


def test_list_versions(setup):
    _, staging, registry, url = pyg.start_gobbler()

    versions = pyg.list_versions("test", "list", registry=registry, url=url)
    assert "v1" in versions
    assert "v2" in versions

    rversions = pyg.list_versions("test", "list", registry=registry, url=url, force_remote=True)
    assert sorted(versions) == sorted(rversions)


def test_list_files(setup):
    _, staging, registry, url = pyg.start_gobbler()

    files = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url))
    assert files == sorted(["..summary", "..manifest", "foo", "whee/blah", "whee2"])
    rfiles = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, force_remote=True))
    assert files == rfiles

    files = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, include_dotdot=False))
    assert files == sorted(["foo", "whee/blah", "whee2"])
    rfiles = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, include_dotdot=False, force_remote=True))
    assert files == rfiles

    files = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, prefix="whee"))
    assert files == sorted(["whee/blah", "whee2"])
    rfiles = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, force_remote=True, prefix="whee"))
    assert files == rfiles

    files = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, prefix="whee/"))
    assert files == sorted(["whee/blah"])
    rfiles = sorted(pyg.list_files("test", "list", "v1", registry=registry, url=url, force_remote=True, prefix="whee/"))
    assert files == rfiles



from typing import Dict, Any
import os
import json
import requests
from . import _utils as ut


def fetch_permissions(project: str, registry: str, url: str, force_remote: bool = False) -> Dict[str, Any]:
    """
    Fetch permissions for a project.

    Args:
        project:
            Name of a project.

        registry:
            Path to the Gobbler registry.

        url:
            URL of the REST API. Only used for remote queries.

        force_remote:
            Whether to force a remote query via ``url``, even if the
            ``registry`` is present on the current filesystem.

    Returns:
        A dictionary containing the permissions for the ``project``.  This
        contains ``owners``, a list of strings with the user IDs of the owners
        of this project; and ``uploaders``, a list of dictionaries where each
        dictionary has the following fields:

        - ``id``, string containing a user ID that is authorized to upload.
        - ``asset`` (optional), string containing the name of the asset that
          the uploader is allowed to upload to. If not present, there is no
          restriction on the uploaded asset name.
        - ``version`` (optional), string containing the name of the version
          that the uploader is allowed to upload to. If not present, there is
          no restriction on the uploaded version name.
        - ``until`` (optional), string containing the expiry date of this
          authorization in Internet Date/Time format. If not provided, the
          authorization does not expire.
        - ``trusted`` (optional), whether the uploader is trusted. If not
          provided, defaults to false.
    """
    if os.path.exists(registry) and not force_remote:
        with open(os.path.join(registry, project, "..permissions"), "r") as f:
            perms = json.load(f)
    else:
        res = requests.get(url + "/fetch/" + project + "/..permissions")
        if res.status_code >= 300:
            raise ut.format_error(res)
        perms = res.json()
    return perms




from typing import Optional, Dict, List
import json
import os
import tempfile
import requests


def format_error(res):
    ctype = res.headers["content-type"]
    if ctype == "application/json":
        info = res.json()
        return requests.HTTPError(res.status_code, info["reason"])
    elif ctype == "text/plain":
        return requests.HTTPError(res.status_code, res.text)
    else:
        return requests.HTTPError(res.status_code)


def dump_request(staging: str, url: str, action: str, payload: Optional[Dict]) -> str:
    if payload is None:
        as_str = character(0)
    else:
        as_str = json.dumps(payload, indent=4)

    prefix = "request-" + action + "-"
    fd, holding_name = tempfile.mkstemp(dir=staging, prefix=prefix)
    with os.fdopen(fd, "w") as handle:
        handle.write(as_str)

    res = requests.post(url + "/new/" + os.path.basename(holding_name))
    if res.status_code >= 300:
        raise format_error(res)

    return res.json()



from . import _utils as ut


def approve_probation(project: str, asset: str, version: str, staging: str, url: str):
    """
    Approve a probational upload of a version of a project's asset.

    Args:
        project:
            The name of the project.

        asset:
            The name of the asset of the ``project``.

        version:
            The name of the version of the ``asset`` to approve.

        staging:
            Path to the staging directory.

        url:
            URL to the Gobbler REST API.
    """
    ut.dump_request(staging, url, "approve_probation", { "project": project, "asset": asset, "version": version })



from typing import List
from .list_projects import list_registry_directories


def list_assets(project: str, registry: str, url: str, force_remote: bool = False) -> List[str]:
    """
    List all assets in a project.

    Args:
        project:
            The name of the project.

        registry:
            Path to the registry.

        url:
            URL to the Gobbler REST API. Only used for remote access.

        force_remote:
            Whether to force remote access via the API, even if ``registry`` is
            on the same filesystem as the caller. 

    Returns:
        List of strings containing the project names.
    """
    return list_registry_directories(project, registry, url, force_remote)



import pygobbler as pyg
import datetime


def test_create_project():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.create_project(
        project="test-create",
        owners=["LTLA", "jkanche"],
        uploaders=[{ "id": "lawremi" }, { "id": "PeteHaitch", "until": datetime.datetime.utcnow().isoformat("T") + "Z" }],
        staging=staging,
        url=url
    )

    perms = pyg.fetch_permissions("test-create", registry, url=url)
    assert perms["owners"] == [ "LTLA", "jkanche" ]
    assert perms["uploaders"][0]["id"] == "lawremi"
    assert perms["uploaders"][1]["id"] == "PeteHaitch"



from typing import Optional
import os
import json
import requests
from . import _utils as ut


def fetch_usage(project: str, registry: str, url: str, force_remote: bool = False) -> Optional[str]:
    """
    Fetch the disk usage of a project.

    Args:
        project:
            Name of a project.

        registry:
            Path to the Gobbler registry.

        url:
            URL of the REST API.

        force_remote:
            Whether to force a remote query via ``url``, even if the
            ``registry`` is present on the current filesystem.

    Returns:
        The current usage (in bytes) for this project.
    """
    if os.path.exists(registry) and not force_remote:
        proposed = os.path.join(registry, project, "..usage")
        with open(proposed, "r") as handle:
            used = json.load(handle)
        return used["total"]

    res = requests.get(url + "/fetch/" + project + "/..usage")
    if res.status_code >= 300:
        raise ut.format_error(res)
    body = res.json()
    return body["total"]



import pygobbler as pyg
import pytest
import os
import time
import tempfile


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)

    tmp = tempfile.mkdtemp()
    pyg.upload_directory(project="test", asset="latest", version="v1", directory=tmp, staging=staging, url=url)
    time.sleep(1.1)
    pyg.upload_directory(project="test", asset="latest", version="v2", directory=tmp, staging=staging, url=url)
    time.sleep(1.1)
    pyg.upload_directory(project="test", asset="latest", version="v3", directory=tmp, staging=staging, url=url)


def test_refresh_latest(setup):
    _, staging, registry, url = pyg.start_gobbler()

    assert pyg.fetch_latest("test", "latest", registry=registry, url=url) == "v3"

    os.unlink(os.path.join(registry, "test", "latest", "..latest"))
    assert pyg.fetch_latest("test", "latest", registry=registry, url=url) is None 

    v = pyg.refresh_latest("test", "latest", staging=staging, url=url)
    assert v == "v3"
    assert pyg.fetch_latest("test", "latest", registry=registry, url=url) == "v3"



import pygobbler as pyg


def test_set_permissions():
    _, staging, registry, url = pyg.start_gobbler()
    pyg.remove_project("test-perms", staging=staging, url=url)
    pyg.create_project("test-perms", staging=staging, url=url, owners=["LTLA"])

    until = "2022-02-02T02:20:02.02Z"
    pyg.set_permissions("test-perms",
        owners=["jkanche"], 
        uploaders=[ { "id": "lawremi", "until": until } ],
        staging=staging,
        url=url,
        registry=registry
    )

    perms = pyg.fetch_permissions("test-perms", registry=registry, url=url)
    assert perms["owners"] == [ "LTLA", "jkanche" ]
    assert len(perms["uploaders"]) == 1
    assert perms["uploaders"][0]["id"] == "lawremi"
    assert perms["uploaders"][0]["until"] == until

    # Checking uploader appending, while also checking owners=NULL.
    pyg.set_permissions("test-perms", uploaders=[ { "id": "ArtifactDB-bot", "trusted": True } ], staging=staging, url=url, registry=registry)
    perms = pyg.fetch_permissions("test-perms", registry=registry, url=url)
    assert perms["owners"] == [ "LTLA", "jkanche" ]
    assert len(perms["uploaders"]) == 2
    assert perms["uploaders"][0]["id"] == "lawremi"
    assert perms["uploaders"][1]["id"] == "ArtifactDB-bot"
    assert perms["uploaders"][1]["trusted"]

    # Checking union of owners, and also that uploaders=NULL works.
    pyg.set_permissions("test-perms", owners=[ "PeteHaitch", "LTLA" ], staging=staging, url=url, registry=registry)
    perms = pyg.fetch_permissions("test-perms", registry=registry, url=url)
    assert perms["owners"] == [ "LTLA", "jkanche", "PeteHaitch" ]
    assert len(perms["uploaders"]) == 2

    # Resetting the owners back.
    pyg.set_permissions("test-perms", owners=[ "LTLA" ], append=False, staging=staging, url=url, registry=registry)
    perms = pyg.fetch_permissions("test-perms", registry=registry, url=url)
    assert perms["owners"] == [ "LTLA" ]
    assert len(perms["uploaders"]) == 2

    # Now resetting the uploaders. 
    pyg.set_permissions("test-perms", uploaders=[], append=False, staging=staging, url=url, registry=registry)
    perms = pyg.fetch_permissions("test-perms", registry=registry, url=url)
    assert perms["owners"] == [ "LTLA" ]
    assert len(perms["uploaders"]) == 0



import pygobbler as pyg
import tempfile
import os
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()
    pyg.remove_project("test", staging=staging, url=url)
    pyg.create_project("test", staging=staging, url=url)

    src = pyg.allocate_upload_directory(staging)
    with open(os.path.join(src, "foo"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(src, "whee"))
    with open(os.path.join(src, "whee", "blah"), "w") as f:
        f.write("stuff")

    pyg.upload_directory("test", "clone", "v1", src, staging=staging, url=url)


def test_clone_version(setup):
    _, staging, registry, url = pyg.start_gobbler()

    dest = tempfile.mkdtemp()
    out = pyg.clone_version("test", "clone", "v1", dest, registry=registry)

    fpath = os.path.join(dest, "foo")
    with open(fpath, "r") as handle:
        assert handle.read() == "BAR"
    assert os.path.exists(os.readlink(fpath))

    fpath = os.path.join(dest, "whee", "blah")
    with open(fpath, "r") as handle:
        assert handle.read() == "stuff"
    assert os.path.exists(os.readlink(fpath))



from . import _utils as ut


def refresh_usage(project: str, staging: str, url:str) -> int:
    """
    Recompute the quota usage of a project. This is useful on rare occasions
    where multiple simultaneous uploads cause the usage calculations to be out
    of sync.

    Args:
        project:
            Name of the project.

        staging:
            Path to the staging directory.

        url:
            URL of the gobbler REST API.

    Returns:
        Total quota usage of this project, in bytes.
    """
    resp = ut.dump_request(staging, url, "refresh_usage", { "project": project })
    return resp["total"]



from typing import Optional, Tuple
import tempfile
import requests
import os
from . import _utils as ut

test_staging = None
test_registry = None
test_port = None
test_process = None

def start_gobbler(staging: Optional[str] = None, registry: Optional[str] = None, port: Optional[int] = None, wait: float = 1, version: str = "0.3.2", overwrite: bool = False) -> Tuple[bool, str, str, str]:
    """
    Start a test Gobbler service.

    Args:
        registry: 
            Path to a registry directory. If None, a temporary directory is
            automatically created.

        staging: 
            Path to a registry directory. If None, a temporary directory is
            automatically created.

        port:
            Port number for the Gobbler API to receive requests. If None, an
            open port is automatically chosen.

        wait:
            Number of seconds to wait for the service to initialize before use.

        version:
            Version of the service to run.

        overwrite:
            Whether to overwrite the existing Gobbler binary.

    Returns:
        A tuple indicating whether a new test service was created (or an
        existing instance was re-used), the path to the staging directory, the
        path to the registry, and the chosen URL. If a service is already
        running, this function is a no-op and the configuration details of the
        existing service will be returned.
    """
    global test_staging
    global test_registry
    global test_process
    global test_port

    if test_process is not None:
        return False, test_staging, test_registry, "http://0.0.0.0:" + str(test_port)

    exe = _acquire_gobbler_binary(version, overwrite)
    _initialize_gobbler_process(exe, staging, registry, port)

    import time
    time.sleep(wait) # give it some time to spin up.
    return True, test_staging, test_registry, "http://0.0.0.0:" + str(test_port)


def _acquire_gobbler_binary(version: str, overwrite: bool):
    import platform
    sysname = platform.system()
    if sysname == "Darwin":
        OS = "darwin"
    elif sysname == "Linux":
        OS = "linux"
    else:
        raise ValueError("unsupported operating system '" + sysname + "'")

    sysmachine = platform.machine()
    if sysmachine == "arm64":
        arch = "arm64"
    elif sysmachine == "x86_64":
        arch = "amd64"
    else:
        raise ValueError("unsupported architecture '" + sysmachine + "'")

    import appdirs
    cache = appdirs.user_data_dir("gobbler", "aaron")
    desired = "gobbler-" + OS + "-" + arch
    exe = os.path.join(cache, desired + "-" + version)

    if not os.path.exists(exe) or overwrite:
        import shutil
        url = "https://github.com/ArtifactDB/gobbler/releases/download/" + version + "/" + desired

        os.makedirs(cache, exist_ok=True)
        tmp = exe + ".tmp"
        with requests.get(url, stream=True) as r:
            if r.status_code >= 300:
                raise ut.format_error(r)
            with open(tmp, 'wb') as f:
                shutil.copyfileobj(r.raw, f)

        os.chmod(tmp, 0o755)

        # Using a write-and-rename paradigm to provide some atomicity. Note
        # that renaming doesn't work across different filesystems so in that
        # case we just fall back to copying.
        try:
            shutil.move(tmp, exe)
        except:
            shutil.copy(tmp, exe)

    return exe
   

def _initialize_gobbler_process(exe: str, staging: Optional[str], registry: Optional[str], port: Optional[int]):
    if staging is None:
        staging = tempfile.mkdtemp()
    global test_staging
    test_staging = staging

    if registry is None:
        registry = tempfile.mkdtemp()
    global test_registry
    test_registry = registry

    if port is None:
        import socket
        with socket.socket(socket.AF_INET) as s:
            s.bind(('0.0.0.0', 0))
            port = s.getsockname()[1]
    global test_port
    test_port = port

    import getpass
    user = getpass.getuser()

    import subprocess
    global test_process
    test_process = subprocess.Popen([ exe, "-admin", user, "-registry", registry, "-staging", staging, "-port", str(port) ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    import atexit
    atexit.register(stop_gobbler)
    return


def stop_gobbler():
    """
    Stop any gobbler test service started by :py:func:`~.start_gobbler`. If no
    test service was running, this function is a no-op.
    """
    global test_process 
    if test_process is not None:
        test_process.terminate()
        test_process = None
    return



import pygobbler as pyg
import pytest
import os
import tempfile


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test-usage", staging=staging, url=url)
    pyg.create_project("test-usage", staging=staging, url=url)

    tmp = tempfile.mkdtemp()
    with open(os.path.join(tmp, "blah.txt"), "w") as f:
        f.write("BAR")
    os.mkdir(os.path.join(tmp, "foo"))
    with open(os.path.join(tmp, "foo", "bar.txt"), "w") as f:
        f.write("1 2 3 4 5 6 7 8 9 10")

    pyg.upload_directory(project="test-usage", asset="quota", version="v1", directory=tmp, staging=staging, url=url)


def test_refresh_usage(setup):
    _, staging, registry, url = pyg.start_gobbler()

    assert pyg.fetch_usage("test-usage", registry=registry, url=url) > 0

    with open(os.path.join(registry, "test-usage", "..usage"), "w") as handle:
        handle.write('{ "total": 0 }')
    assert pyg.fetch_usage("test-usage", registry=registry, url=url) == 0

    # Fixing the project usage.
    out = pyg.refresh_usage("test-usage", staging=staging, url=url)
    assert out > 0
    assert pyg.fetch_usage("test-usage", registry=registry, url=url) == out



import pygobbler as pyg
import tempfile
import os
import time
import pytest


@pytest.fixture(scope="module")
def setup():
    _, staging, registry, url = pyg.start_gobbler()

    pyg.remove_project("test-R-remove", staging=staging, url=url)
    pyg.create_project("test-R-remove", staging=staging, url=url)

    src = pyg.allocate_upload_directory(staging)
    pyg.upload_directory("test-R-remove", "sacrifice", "v1", src, staging=staging, url=url)
    time.sleep(1.1) # force timestamps to be different for next versions.
    pyg.upload_directory("test-R-remove", "sacrifice", "v2", src, staging=staging, url=url)


def test_remove_version(setup):
    _, staging, registry, url = pyg.start_gobbler()
    assert os.path.exists(os.path.join(registry, "test-R-remove", "sacrifice", "v2"))
    pyg.remove_version("test-R-remove", "sacrifice", "v2", staging=staging, url=url)
    assert not os.path.exists(os.path.join(registry, "test-R-remove", "sacrifice", "v2"))


def test_remove_asset(setup):
    _, staging, registry, url = pyg.start_gobbler()
    assert os.path.exists(os.path.join(registry, "test-R-remove", "sacrifice"))
    pyg.remove_asset("test-R-remove", "sacrifice", staging=staging, url=url)
    assert not os.path.exists(os.path.join(registry, "test-R-remove", "sacrifice"))


def test_remove_project(setup):
    _, staging, registry, url = pyg.start_gobbler()
    assert os.path.exists(os.path.join(registry, "test-R-remove"))
    pyg.remove_project("test-R-remove", staging=staging, url=url)
    assert not os.path.exists(os.path.join(registry, "test-R-remove"))



from typing import Optional, List
from . import _utils as ut


def create_project(project: str, staging: str, url: str, owners: Optional[List] = None, uploaders: Optional[List] = None):
    """
    Create a new project in the registry.

    Args:
        project:
            Name of the project to create.

        staging:
            Path to the staging directory.

        url:
            URL for the Gobbler REST API.

        owners:
            List of user IDs of the owners of this project. If not provided,
            the current user will be set as the sole owner.

        uploaders:
            List specifying the authorized uploaders for this project. See the
            ``uploaders`` field in :py:func:`~.fetch_permissions` return value
            for the expected format. 

    Returns:
        On success, the requested project is created in the registry.
    """

    req = { "project": project }

    permissions = {}
    if owners is not None:
        permissions["owners"] = owners

    if uploaders is not None:
        permissions["uploaders"] = uploaders

    if len(permissions):
        req["permissions"] = permissions

    ut.dump_request(staging, url, "create_project", req)
    return



from typing import List
import requests
import os
from . import _utils as ut


def list_projects(registry: str, url: str, force_remote: bool = False) -> List[str]:
    """
    List all projects in the registry.

    Args:
        registry:
            Path to the registry.

        url:
            URL to the Gobbler REST API. Only used for remote access.

        force_remote:
            Whether to force remote access via the API, even if ``registry`` is
            on the same filesystem as the caller. 

    Returns:
        List of strings containing the project names.
    """
    listed = list_registry_directories(".", registry, url, force_remote)

    # Remove the ..logs directory.
    output = []
    for x in listed:
        if not x.startswith(".."):
            output.append(x)

    return output


def list_registry_directories(path: str, registry: str, url: str, force_remote: bool) -> List[str]:
    if force_remote and os.path.exists(registry):
        output = []
        full = os.path.join(registry, path)
        for x in os.listdir(full):
            if os.path.isdir(os.path.join(full, x)):
                output.append(x)
        return output

    else:
        url += "/list"
        if path != ".":
            import urllib
            url += "?path=" + urllib.parse.quote_plus(path)

        req = requests.get(url)
        if req.status_code >= 300:
            raise ut.format_error(req)

        listing = req.json()
        output = []
        for x in listing:
            if x.endswith("/"):
                output.append(x[:-1])

        return output



from typing import Optional, List, Dict
from . import _utils as ut
from . import fetch_permissions


def set_permissions(project: str, registry: str, staging: str, url: str, owners: Optional[List] = None, uploaders: Optional[Dict] = None, append: bool = True):
    """
    Set the owner and uploader permissions for a project.

    Args:
        project:
            Name of the project.

        registry:
            Path to the Gobbler registry.

        staging:
            Path to the staging directory.

        url:
            URL of the REST API.

        owners:
            List of user IDs for owners of this project. If None, no change is
            made to the existing owners in the project permissions.

        uploaders:
            List of dictionaries specifying the authorized uploaders for this
            project.  See the ``uploaders`` field in the return value of
            :py:func:`~.fetch_permissions` for  the expected format. If None,
            no change is made to the existing uploaders.

        append:
            Whether ``owners`` and ``uploaders`` should be appended to the
            existing owners and uploaders, respectively. If False, the
            ``owners`` and ``uploaders`` are used to replace the existing
            values in the project permissions.
    """
    perms = {}

    if append:
        oldperms = fetch_permissions(project, registry=registry, url=url)
        if owners is not None:
            oldset = set(oldperms["owners"])
            perms["owners"] = oldperms["owners"] + list(filter(lambda x : x not in oldset, owners))
        if uploaders is not None:
            perms["uploaders"] = oldperms["uploaders"] + uploaders
    else:
        if owners is not None:
            perms["owners"] = owners
        if uploaders is not None:
            perms["uploaders"] = uploaders

    ut.dump_request(staging, url, "set_permissions", { "project": project, "permissions": perms })



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .start_gobbler import start_gobbler, stop_gobbler
from .create_project import create_project
from .fetch_permissions import fetch_permissions
from .fetch_latest import fetch_latest
from .fetch_summary import fetch_summary
from .fetch_manifest import fetch_manifest
from .fetch_usage import fetch_usage
from .fetch_directory import fetch_directory
from .allocate_upload_directory import allocate_upload_directory
from .upload_directory import upload_directory
from .list_projects import list_projects
from .list_assets import list_assets
from .list_versions import list_versions
from .list_files import list_files
from .remove_project import remove_project
from .remove_asset import remove_asset
from .remove_version import remove_version
from .clone_version import clone_version
from .service_info import service_info
from .version_path import version_path
from .refresh_latest import refresh_latest
from .refresh_usage import refresh_usage
from .approve_probation import approve_probation
from .reject_probation import reject_probation
from .set_permissions import set_permissions
